/* 	Copyright John D. "smeezekitty" 2017
	
	FreeNBTV is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeNBTV is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeNBTV.  If not, see <http://www.gnu.org/licenses/>.
*/

/* This file handles the output*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <portaudio.h>
#include "nbtv_format.h"
#include "input.h"
#include "gui.h"
#include "nbtv_gen.h"
#include "output.h"
#include "util.h"

uint8_t output_samples[1<<20];
uint8_t active_output[1<<18];
uint8_t ob[4][1<<20];
struct _output output;
int audio_lock;
int buf_ptr=0;
extern int frame_len;
extern char repeat_frame;
int updated_buf;
PaStream *stream;
double output_timestamp;
void audio_fill(void *udata, uint8_t *stream, int len){
	int l,p=0;
	//memcpy(stream, buffer, l);
	/*if(len < frame_len){
		printf("Audio dis\n");
	}*/
	while(len){
	//	printf("frame_len = %d buf_ptr = %d len=%d\n", frame_len, buf_ptr, len);
		audio_lock = 1;
		if(buf_ptr >= (frame_len*(output.bps/8)*output.num_channels)){
			buf_ptr = 0;
			if(updated_buf){
			//	memcpy(active_output, output_samples, (frame_len*(output.bps/8)*output.num_channels));
			//	updated_buf = 0;
			}
		}
		l = min((frame_len*(output.bps/8)*output.num_channels) - buf_ptr, len);
		//printf("p=%d buf_ptr=%d l=%d\n", p, buf_ptr, l);
		memcpy(stream+p, output_samples+buf_ptr, l);
		len -= l;
		p+=l;
		buf_ptr += l;
		audio_lock = 0;
	}
}
/*Portaudio callback*/
static int audio_fill2(const void *i, void *o, unsigned long len, const PaStreamCallbackTimeInfo *time_info, PaStreamCallbackFlags status, void *u){
	int l,p=0;
	while(len){
	//	printf("frame_len = %d buf_ptr = %d len=%d\n", frame_len, buf_ptr, len);
		audio_lock = 1;
		if(buf_ptr >= (frame_len*(output.bps/8)*output.num_channels)){
			buf_ptr = 0;
			if(updated_buf){
			//	memcpy(active_output, output_samples, (frame_len*(output.bps/8)*output.num_channels));
			//	updated_buf = 0;
			}
		}
		l = min((frame_len*(output.bps/8)*output.num_channels) - buf_ptr, len * (output.bps/8)*output.num_channels);
		//frames = min((frame_len - buf_ptr, len);
		//printf("p=%d buf_ptr=%d l=%d\n", p, buf_ptr, l);
		memcpy(o+p, output_samples+buf_ptr, l);
		len -= (l/((output.bps/8)*output.num_channels));
		p+=l;
		buf_ptr += l;
		audio_lock = 0;
	}
	return 0;
}

void nbtv_output(float channels[4][1<<17], uint8_t *key, int sc, struct NBTV_format *f, struct _output *of){
	int seq;
	int channel;
	int channel_seq[4];
	int mc_count = 0;
	int i;
	uint32_t sample;
	uint32_t half_scale = (1<<(of->bps-1))-1;
	int sample_index;
	double gain_factor;
	uint8_t *o_samples;
	uint8_t main_channel_flag = 0;
	if(repeat_frame){
		goto write_only;
	}
for(channel = 0;channel < of->num_channels;channel++){
	debug("Processing channel %d\n", channel);
		/*	if(of->channels[channel].format == OUTPUT_FORMAT_MAIN){
				seq = of->channels[channel].main_channel_num;
				main_channel_flag = 1;
				printf("Main channel. seq=%d\n", seq);
			} else if(of->channels[channel].format == 0){
				continue;
			} else main_channel_flag = 0;*/
			main_channel_flag = 1;
			seq = channel;
			struct NBTV_channel *ch = &(f->channels)[channel];
			float *samples = (float*)&channels[channel];
			gain_factor = 1.0;
			if(ch->channel_gain > 0){
				gain_factor += ch->channel_gain/50.0;
			} else {
				gain_factor += ch->channel_gain/125.0;
			}
			for(i = 0;i < sc;i++){
				if(!key[i]){
					if(ch->invert_video)samples[i] = 1.0-samples[i];
					samples[i] *= (double)ch->video_level/100.0;
					samples[i] += (double)(ch->video_dc_off/100.0);
				} 
				samples[i] *= gain_factor;
				if(ch->invert_channel){samples[i] =- samples[i];}
				if(samples[i] > 1.0){samples[i] = 1.0;}
				if(samples[i] < -1.0){samples[i] = -1.0;}
				/*Performs low pass filtering*/
				sample = half_scale*!of->signed_samples + half_scale * (samples[i] * 0.6 + (i? samples[i-1]: samples[i]) * 0.20 + ((i+1<sc)?samples[i+1]: samples[i]) * 0.20);
				if(main_channel_flag){
					o_samples = (uint8_t*)&output_samples;
					sample_index = i * of->num_channels + seq;
				} else {
					o_samples = (uint8_t*)&ob[channel];
					sample_index = i;
				}
				switch(of->bps){
					case 4:
						o_samples[sample_index>>1] = sample << ((!(sample_index&1))<<2);
					case 8:
						if(of->signed_samples){
							((int8_t*)o_samples)[sample_index] = sample;
						} else {
							o_samples[sample_index] = sample;
						}
					break;
					case 16:
						if(of->signed_samples){
							((int16_t*)o_samples)[sample_index] = sample;
						} else {
							((uint16_t*)o_samples)[sample_index] = sample;
						}
					break;
					case 32:
						if(of->signed_samples){
							((int32_t*)o_samples)[sample_index] = sample;
						} else {
							((uint32_t*)o_samples)[sample_index] = sample;
						}
						break;
					default:
						error("nbtv_output: unknown bit depth\n");
						return;
			}
		}
		
	}
	write_only:
	if(of->format != OUTPUT_FORMAT_SC){
		//fwrite(o_samples, 1, sc, of->file);
		audio_file_write_block(of, o_samples, sc);
	}
	output_timestamp += 1000.0/f->frame_rate;
}
void setup_output(struct _output *o){
//	SDL_AudioSpec af, actual;
	if(o->format != OUTPUT_FORMAT_SC){
		return;
	}
	if(o->bps == 4){
		o->bps = 8; //The sound card doesn't support 4 BPS
	}
	/*af.freq = o->sample_rate;
	af.format = (o->bps & 0xFF) | o->signed_samples << 15;
	af.channels = o->num_channels;
	af.samples = 2048;
	af.callback = audio_fill;
	printf("CAlling open audio with sr=%d format=%d ch=%d\n", af.freq, af.format, af.channels);
	SDL_OpenAudio(&af, &actual);
	if(actual.freq != o->sample_rate){
		printf("Sample rate couldn't be set to %d. Set to %d instead\n", o->sample_rate, actual.freq);
		o->sample_rate = actual.freq;
	}
	if(actual.format & 0xFF != o->bps){
		printf("Bits per sample couldn't be set to %d. Set to %d instead\n", o->bps, actual.format & 0xFF);
		o->bps = actual.format & 0xFF;
	}
	if(!(actual.format & (1<<15)) == o->signed_samples){
		printf("%s samples couldn't be set. Set %s samples instead\n", o->signed_samples? "Signed": "Unsigned", o->signed_samples? "unsigned": "signed");
		o->signed_samples = !!(actual.format & (1<<15));
	}
	SDL_Delay(250);
	printf("Calling SDL_CloseAudio()\n");
	SDL_CloseAudio();*/
//	if(!stream){
//	}

}
void start_output(struct _output *o){
	//SDL_AudioSpec af, actual;
	if(o->format != OUTPUT_FORMAT_SC){
		o->file = fopen(o->main_filename, "wb");
		if(o->format == OUTPUT_FORMAT_WAVE){
			ms_wave_write_start(o);
		}
		gui_format_disable_window();
		return;
	} 
/*	af.freq = o->sample_rate;
	af.format = (o->bps & 0xFF) | o->signed_samples << 15;
	af.channels = o->num_channels;
	af.samples = 1024;
	af.callback = audio_fill;
	printf("CAlling open audio with sr=%d format=%d ch=%d\n", af.freq, af.format, af.channels);
	if(SDL_OpenAudio(&af, &actual)){
		printf("Error opening audio\n");
		return;
	}
	SDL_PauseAudio(0);*/
	int fmt;
	switch(o->bps){
		//case 4:
		//	o->sample_rate /= 2;
		default:
		case 8:
			fmt = o->signed_samples? paInt8: paUInt8;
			break;
		case 16:
			fmt = paInt16;
			break;
		case 32:
			fmt = paInt32;
	}
	/*TODO: Add device selection*/
	Pa_OpenDefaultStream(&stream, 0, o->num_channels, fmt, o->sample_rate, 0, audio_fill2, NULL);
	//o->sample_rate *= 1 + (o->bps == 4);
	Pa_StartStream(stream);
}
void stop_output(){
	if(output.format != OUTPUT_FORMAT_SC){
		gui_format_enable_window();
		audio_file_end(&output);
		return;
	}
	//SDL_PauseAudio(1);
	//SDL_CloseAudio();
	Pa_StopStream(stream);
	Pa_CloseStream(stream);
	stream = NULL;
}